#include "WPILib.h"
#include "Camera.h"
#include "IOmap.h"
#include "ThunderDrive.h"
#include "RPS.h"
#include <math.h>

RobotPositioningSystem::RobotPositioningSystem(Camera* c) {
	/* TODO need to know what channels are being used! */
	accel = new ADXL345_I2C(0);
	cam = c;
	timer = new Timer();
	update_counter = 0;
	// initialize variables and arrays
	for(int i=0;i<6;i++) {
		dxs[i]=dys[i]=0;
	}
	// this is rather ugly... oh well, it calculates easier
	// while avoiding hard-coded values...
#ifdef SPLIT_WEIGHTS
	weights[0] = WT_D_E*WT_A_E; weights[1] = WT_D_E*WT_A_G;
	weights[2] = WT_D_E*WT_A_A;	weights[3] = WT_D_A*WT_A_E;
	weights[4] = WT_D_A*WT_A_G;	weights[5] = WT_D_A*WT_A_A;
#else
	weights[0] = WT_EE; weights[1] = WT_EG;	weights[2] = WT_EA;
	weights[3] = WT_AE;	weights[4] = WT_AG;	weights[5] = WT_AA;
#endif
}

RobotPositioningSystem::~RobotPositioningSystem() {
	// TODO what do we need to clean up?
	delete timer;
	delete accel;
}

void RobotPositioningSystem::SetXYPosAngle(float x, float y, float theta) {
	xpos = x;
	ypos = y;
	angle = theta;
}
float RobotPositioningSystem::GetXPos() {
	return xpos;
}
float RobotPositioningSystem::GetYPos() {
	return ypos;
}
float RobotPositioningSystem::GetAngle() {
	return angle;
}

			/* THE BIG ONE! */
void RobotPositioningSystem::UpdatePosition() {
	/* TODO can RPS be declared broken? is it even necessary? */ 
	// lots and lots and lots of math!!!
	// set to 1000 as default; if still equal 1000, we know camera data failed
	float x_cam=1000, y_cam=1000;
	/* blah blah blah....; then we insert this: */
	// this is either really compact and elegant, or awful practice. I don't know which.
	if(++update_counter >= UPDATES_BEFORE_NEW_IMAGE) {
		// time to update the camera, and do all that funky math
		cam->GetNewImage();
		// proceed only if we got usable info from the camera!
		if(cam->TargetIdentified()) {
			update_counter = 0; // reset the count so we can start over		
			// and run the numbers
			/* TODO can we use camera to determine robot angle ???! */
			GetXYCamera(&x_cam,&y_cam);
		}
	}
	
	// and back to the general math...
	// get various differences between now and last time
	float d_t = timer->Get() - time;
	int d_lticks = ThunderDrive::getInstance()->getEncoder(false) - lticks;
	int d_rticks = ThunderDrive::getInstance()->getEncoder(false) - rticks;
	float g_d_angle = GetGyroAngle() - g_angle;
	double d_a_x = GetAX() - a_x;
	double d_a_y = GetAY() - a_y;
	
	/* This is where the magic happens - 
	 * (see math documentation for further details) */
	
	// difference in left side and right side position
	// (if we trust the encoders)
	float d_l = d_lticks * INCHES_PER_TICK;
	float d_r = d_rticks * INCHES_PER_TICK;
	/** part one of the magic: angle turned (in radians) when one side
	 * travels somewhat farther than the other equals (d_l-d_r)/drivewidth:
	 * positive when turning right and negative when turning left!!! */
	float e_d_angle = (d_l - d_r) / DRIVEWIDTH;
	/** and, from the same model, the distance the robot has traveled is just
	 * the average of the left and right distances! (if your time delta is 
	 * small enough, which it is in this case) */
	float e_d_d = (d_l + d_r) / 2;
	/** integrate change in acceleration to find change in speed and
	 * then change in distance */
	// first in robot's direction (assuming approx. linear path over minimal time)
	float d_v_x = (a_x + d_a_x/2) * d_t;
	float d_d_x = (v_x + d_v_x/2) * d_t;
	// then perpendicular
	float d_v_y = (a_y + d_a_y/2) * d_t;
	float d_d_y = (v_y + d_v_y/2) * d_t;
	/** and then the change in angle is.... d_d_y / d_d_x ?!!!!!! */
	float a_d_angle = d_d_y / d_d_x;
	// standard pythagorean stuff, this:
	float a_d_d = sqrt(d_d_x*d_d_x + d_d_y*d_d_y);
	// note: d_d_x is NOT the change in *field-coordinate* X position; ditto for Y
	
	/**** This is the deepest part of the magic ****/
	// various combinations of distance and angle
	// both from encoder
	dxs[0] = e_d_d*cos(e_d_angle);
	dys[0] = e_d_d*sin(e_d_angle);
	// encoder distance, gyro angle
	dxs[1] = e_d_d*cos(g_d_angle);
	dys[1] = e_d_d*sin(g_d_angle);
	// encoder distance, accel. angle
	dxs[2] = e_d_d*cos(a_d_angle);
	dys[2] = e_d_d*sin(a_d_angle);
	// accel. distance, encoder angle
	dxs[3] = a_d_d*cos(e_d_angle);
	dys[3] = a_d_d*sin(e_d_angle);
	// accel. distance, gyro angle
	dxs[4] = a_d_d*cos(g_d_angle);
	dys[4] = a_d_d*sin(g_d_angle);
	// both from accel.
	dxs[5] = a_d_d*cos(a_d_angle);
	dys[5] = a_d_d*sin(a_d_angle);
					
	// Now reconcile all those different values -- how? Via weighted average!!
	// The magic is all in the #defined weights!
	float d_x=0, d_y=0;
	for(int i=0;i<6;i++) {
		d_x += dxs[i]*weights[i];
		d_y += dys[i]*weights[i];
	}
	/**** End of the deep magic ****/
	// now just add the difference to get new position
	xpos += d_x;
	ypos += d_y;
	
	// but if we got a camera picture, take its data into account as well
	// (cam gives absolute positions, so we can't put it into the weighted
	//  average of *changes in position* above)
	if(x_cam!=1000 && y_cam!=1000) {
		xpos = xpos*WT_NONCAM + x_cam*WT_CAM;
		ypos = ypos*WT_NONCAM + y_cam*WT_CAM;
	}
	
	// robot's angle is treated similarly to position, but with separate
	// weights and no camera interference; some deep magic here as well
	angle = angle + e_d_angle*WT_A_E + g_d_angle*WT_A_G + a_d_angle*WT_A_A;
	/* End of the magic */
	
	// final point of order: update everything by adding back the
	// differences we subtracted in the first place (trollface.jpg)
	time += d_t;
	lticks += d_lticks;
	rticks += d_rticks;
	g_angle += g_d_angle;
	a_x += d_a_x;
	a_y += d_a_y;
	v_x += d_v_x;
	v_y += d_v_y;
	return;
}

/* --- private functions  --- */
void RobotPositioningSystem::GetXYCamera(float* x, float* y) {
	float z = cam->TargetCenterHeight();
	
	/** see PositionDetermination and Camera.h for math and constant info, respectively */
	// angle from camera to target--in the vertical!!
	float alpha = asin(z*cam->HeightOfTarget()*TWO_FH0) / 2;
	// horizontal distance from camera to target:
	float d_h = z/tan(alpha);
	// various intermediate values
	float beta = acos(z/sin(alpha)*cam->WidthOfTarget()/FW0);
	float sinbetah = sin(beta)/cos(alpha);
	// and the final x/y position calculated (d_h terms) is relative to target
	// so add relative distances to target position to get robot position
	*x = MAX_X - d_h*cos(asin(sinbetah)); // x distance is from wall (at max_x)
	// y is distance from target-specific ypos; also, can be positive or negative
	// depending on which side of it we're on (accounted for with TargetTilt())
	/* TODO what if TT() is zero? */
	*y = cam->TargetCenterYPos() + d_h*sinbetah*cam->TargetTilt(); 
	/* TODO we're assuming we're looking at our alliance's targets (opposite our Alliance
	 * Station... but how do we know which side of the field we're looking at?? */
	/* TODO correct for camera position vs. robot center? */
	
}

// quick wrappers with a little math in them
float RobotPositioningSystem::GetGyroAngle() {
	return ThunderDrive::getInstance()->getGyro()*PI_180;
}
double RobotPositioningSystem::GetAX() {
	return accel->GetAcceleration(ADXL345_I2C::kAxis_X)*G_TO_IN_S2;
}
double RobotPositioningSystem::GetAY() {
	return accel->GetAcceleration(ADXL345_I2C::kAxis_Y)*G_TO_IN_S2;
}
