#include "WPILib.h"
#include "Storage.h"
#include "Shooter.h"
#include "RPS.h"
#include "Camera.h"
#include "SmartShooter.h"
#include <math.h>
// do we want to make ABSOLUTELY CERTAIN auto-targeting won't break ANYTHING?
#define KILL_TARGETING  
// do we want to make ABSOLUTELY CERTAIN that the RPS can't break ANYTHING?
#define KILL_RPS

// static lookup tables; TODO put in BETTER values!!!
// (actual angles are +50..-39)
float SmartShooter::rough_speed_lookup_table[RSLT_SIZE] = 
	{ // +10    +9    +8    +7    +6    +5    +4    +3    +2    +1 
		.612, .613, .613, .613, .613, .614, .614, .614, .615, .615,  //40
		.616, .616, .617, .618, .619, .620, .621, .623, .625, .628,  //30
		.631, .635, .639, .644, .652, .670, .700, .740, .780, .830,  //20
		.880, .935, .990, 1.00, 1.00, 1.00, 1.00, 1.00, 1.00, 1.00,  //10
		/*don't really bother below 10 degrees above horizontal...*/
		0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00,  //0
		0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00,  //-10
		0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00,  //-20
		0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00,  //-30
		0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00   //-40
	}; // NOTE: MUST HAVE RSLT_SIZE VALUES!
float SmartShooter::speed_lookup_table[3][SLT_DSIZE] =
	{ { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
		0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
		0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0
	  }, // height = 0 (bottom basket)
	  { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
	  	0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
	  	0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0
	  }, // height = 1 (middle baskets)
	  { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
	  	0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
	  	0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0
	  } // height = 2 (top baskets)
	}; // NOTE: MUST HAVE 3*SLT_DSIZE VALUES!

SmartShooter::SmartShooter(Storage *s) : Shooter(s) {
	cam = new Camera();
#ifdef KILL_RPS
	rps = NULL;
#else
	rps = new RobotPositioningSystem(cam);
#endif
//	shot_requested = false; ** deprecated
	targeting = false; /* should be set true by ThunderAuto if hybrid desires targeting! */
	positioned = false;
	target_hoop = Camera::none;
	mode = camera_only;
	camera_tries = 0;
}

SmartShooter::~SmartShooter() {
	// TODO what cleanup to put here?
	delete cam;
	delete rps;
}

void SmartShooter::processShooter() {
	// FUN TIME! ROBOT BEGINS TO DESTROY ITSELF!!!
#ifndef KILL_RPS // KILL_RPS stops RPS cold by never allowing it to be called
	// ALWAYS update your position! no matter what happens
	/* TODO would this actually be better off in the main robot loop???? */
	rps->UpdatePosition();
#endif
	
	// not targeting? skip this whole if-block and just go on
	// to superclass method!
#ifndef KILL_TARGETING
	// calculations occur only if a) we are targeting; b) we haven't
	// found our position yet; c) we haven't disabled ourselves with off mode :)
	// otherwise we can just skip it!
    if(targeting && !positioned && mode!=off) {
		if(mode!=rps_only) {
			cam->GetNewImage(); // force a new image if we're using camera
			if(!cam->TargetFound() || 
					(target_hoop==Camera::none && !cam->TargetIdentified())) {
				// either no Target, or we've found one but can't identify it
				//   AND the drivers haven't told us the target ID either
				positioned = false; // don't know adequately where we are!
				camera_tries++; // one more failed attempt
			} else {					
				camera_tries = 0;
				// we're positioned for shot if target is adequately centered
				positioned = cam->TargetCentered();
			}
		} else positioned = (target_hoop!=Camera::none);
		// ^^ if we're just using RPS, we trust its position implicitly, but 
		// only if the drivers have told us which hoop to shoot at!!
		
		// check for problems; if we have any, give up (and notify controls)
		// camera can't find it:
		if(camera_tries > CAMERA_TRIES_LIMIT) failure=true;
			// OR: should we switch automatically to rps_only?
		// stuff broke while we were still targeting! horrors:
		if(wheels_broken || broken_turret) failure=true;
		/* TODO should targeting have a time limit (i.e. if we can't
		 *      get on target within X runs, give up)? or should we leave
		 *      the decision to the drivers? */
		
		if(failure) {
			// TRY to shut down gracefully: stop wheels and turret
			/* TODO is this warranted? could it create more problems than
			 *      it solves? */
			setSpeed(0,0);
			setTurretAngle(getTurretAngle());
		} else { // we can proceed!
			// get the angle and speed required, either to get target more centered
			// (if using camera and not yet positioned) or for the actual shot (if
			//  we are positioned); set speed if not positioned merely to spin up wheels
			// if CalcTDandA fails, everything safely stays at zero
			float angle = 0;
			float distance = 0;
			CalcTargetDistanceAndAngle(&distance,&angle);
			angle = getTurretAngle() + angle;
			// speed calculation!!
			float speed = 0;
			// what's the closest position in the array not higher than
			// the calculated distance?
			int distance_i = (int)((distance-X0)/DX);
			switch(target_hoop) {
			case Camera::top: 
				speed = speed_lookup_table[2][distance_i];
				// linear interpolation! y = dy/dx|x=x_i * (x-x_i)  
			    speed += ((speed_lookup_table[2][distance_i+1]-speed)/DX)
			    		 * (distance-distance_i);
				break;
			case Camera::center_left:
			case Camera::center_right:
			case Camera::center: 
				speed = speed_lookup_table[1][distance_i];
			    speed += ((speed_lookup_table[1][distance_i+1]-speed)/DX)
			    		 * (distance-distance_i);
				break;
			case Camera::bottom:
				speed = speed_lookup_table[0][distance_i];
			    speed += ((speed_lookup_table[0][distance_i+1]-speed)/DX)
			    		 * (distance-distance_i);
				break;
			default: break; // can't set a speed!!
			}
		
			
			// finish up by setting our new targets:
			/* do we need to modify the speeds in any way to create the backspin? TODO */
			setSpeed(speed,speed);
			/* and what do we do if the turret encoder is broken? */
			setTurretAngle(angle);
		} // endif [!failure]
    } /* endif(targeting && !positioned && mode!=cheating) */		
#endif	
	// whether or not we just ran targeting stuff, give control
	// to the original version to make everything move!
	Shooter::processShooter();
}

void SmartShooter::setTargeting(bool t) {
	targeting = t;
	if(wheels_broken || broken_turret) {
		// broken encoders make targeting useless; don't even start!
		targeting = false;
	}
	if(!targeting) {
		positioned = false; // we don't care where we are if we're not targeting
		/* TODO check that this won't spam failure=false over processShooter's
		 *      legitimate claims of failure! */
		failure = false; // and we haven't yet tried, so we can't have failed!
	}
}
void SmartShooter::SetTargetMode(TargetMode m) {
	mode = m;
}
void SmartShooter::SetTargetHoop(Camera::TargetIdentity t) {
	target_hoop = t;
}

bool SmartShooter::onTarget() {
	// to be on target, we must be first of all looking for the target, 
    // and have a secure fix on our location!
	return targeting && positioned;
}
bool SmartShooter::targetingFailed() {
	return failure;
}

void SmartShooter::tiltCamera(float angle) {
	if(targeting) return; // don't mess with my camera!
	cam->SetTilt(angle);
}
float SmartShooter::getCamTilt() {
	return cam->GetTilt();
}
bool SmartShooter::camTiltDone() {
	return cam->TiltDone();
}

float SmartShooter::getRoughTargetSpeed() {
	float angle = cam->GetTilt();
	// floor of angle; used as index in rough lookup table
	int index = (int)angle;
	// get fractional angle; will linearly integrate to 
	// fine-tune the target
	angle -= index;
	// we can't have out-of-range errors!!
	if(index<MIN_ANGLE) return 0.0;
	if(index>=(MAX_ANGLE-1)) return 0.0;
	float rslt = rough_speed_lookup_table[index];
	// delta is slope (dx=1) times fractional delta_x
	float delta = (rough_speed_lookup_table[index+1] - rslt)*angle;
	// round delta to hundreds place because that's the max
	// resolution of the target speed display
	int ret = (int)((rslt+delta)*100);
	return (float)(ret/100.0);
	//return rslt;
}

/* --- private functions below this point --- */
void SmartShooter::CalcTargetDistanceAndAngle(float* distance, float* angle) {
	float z=0; // difference in height between camera and hoop
	// first ask drivers for hoop, ignoring any camera input;
	// but if we're not told we're looking at a specific target,
	// hand the question over to the camera
	switch(target_hoop) {
	case Camera::none: break; // no specific target--ask camera
	case Camera::top: z=TOP_HEIGHT; break;
	case Camera::center_left:
	case Camera::center_right:
	case Camera::center: z=CENTER_HEIGHT; break;
	case Camera::bottom: z=BOTTOM_HEIGHT; break;
	}
	
	// TODO do we need to do some additional analysis if tgthoop==center?
	switch (mode) {
	case camera_only:
	case camera_and_rps:
		if(z==0) { //ask cam
			z=cam->TargetCenterHeight();
			if(z==0) return; // cam doesn't know either? give up
		}
		/** see PositionDetermination for gory math details; see Camera.h for
		 *  meanings of #defined constants */
		// angle from camera to target--in the vertical!!
		float alpha = asin(z*cam->HeightOfTarget()*TWO_FH0) / 2;
		// horizontal distance from camera to target:
		float d_h = z/tan(alpha);
		// now, if we're not fully positioned, use an approximation to
		// get closer and spin up the shooter wheels
		float d_x,d_y;
		if(!positioned) {
			*distance = d_h;
			// following is VERY approximate! actual angle would be tan of this value.
			// also note that this equals THD*d_h/F (actual size of THD) divided by d_h
			// if we need to turn turret to left (positive angle), THD will be negative
			*angle = (float)(-cam->TargetHorizDelta()/F)/PI_180;
		} else { // we know where we are; bring out the big guns:
			//calculate the actual desired shooter angle and shot distance!
			float beta = acos(z/sin(alpha)*cam->WidthOfTarget()/FW0);
			float sinbetah = sin(beta)/cos(alpha);
			float betah = asin(sinbetah);
			
			// then: shooter's x and y distance to basket center. See docs
			// y is || target plane (or alliance station wall), x is perpendicular 
			d_x = (d_h-CAM_SHOOTER_DIST)*cos(betah) - T_B_DIST;
			d_y = (d_h-CAM_SHOOTER_DIST)*sinbetah;
			
			// and here are the final values! angle = phi - beta_h: see docs
			// if robot is to right of target, it will have to correct by turning its
			// turret left (positive); TargetTilt() provides the sign correction
			// (if TT()==0, angle should be zero anyway, since we're facing
			//  target straight-on!)
			// TODO does this work automatically for distance (i.e. will d_y be
			//      practically zero) or will we need to do it manually?
			*angle = (atan(d_y/d_x) - betah)/PI_180*cam->TargetTilt();
			*distance = sqrt(d_x*d_x+d_y*d_y);
		}
	    if(mode==camera_only) break;
#ifndef KILL_RPS
	    // mode==camera_and_rps: we're not done yet!
	    d_x = MAX_X - rps->GetXPos();
	    d_y = cam->TargetCenterYPos() - rps->GetYPos();
	    // robot angle+robot center angle to target = turret angle
	    float theta = rps->GetAngle() + atan(d_y/d_x); 
	    float d = sqrt(d_x*d_x+d_y*d_y);
	    // and... yet another weighted average!!!
	    /* TODO is weighted average appropriate here? */
	    *angle = *angle*WT_A_CAM + theta*WT_A_RPS;
	    *distance = *distance*WT_D_CAM + d*WT_D_RPS;
#endif	    
		break;
	case rps_only:
#ifndef KILL_RPS
		/* TODO any way to avoid having to copy this between here and camera_and_rps?
		 * until then this is the definitive version */
		if(z==0) return; // we can't aim if we don't know where we're aiming!
		float dx = MAX_X - rps->GetXPos();
		float dy = cam->TargetCenterYPos() - rps->GetYPos();
		// robot angle+robot center angle to target = turret angle
		*angle = rps->GetAngle() + atan(dy/dx); 
		*distance = sqrt(dx*dx+dy*dy);
		/* TODO correct for shooter displacement from robot center? */
#endif
		break;
	default:
		break; // we shouldn't even be here....
	} // switch(mode)
}
