#ifndef UDPLOGGER_H
#define UDPLOGGER_H
#include "PIDController.h"
#include "PIDOutput.h"
#include "MonoEncoder.h"
#include <sockLib.h>
#include <netinet/in.h>


class UDPLogger {
public:
	UDPLogger(bool stdoutprint = false);
	~UDPLogger();
	
	void init(int port);
	void log(const char *fmt, ...);
	void setEnabled(bool en);
private:
#define UDPLOGGER_BUFSIZE 65536
	char buf[UDPLOGGER_BUFSIZE];
	int fd;
	bool stdoutprint;
	bool enabled;
	struct sockaddr_in saddr;
};

/* 
 * Wraps a PIDOutput, logging during each call to PIDWrite() out the specified
 * UDPLogger.  Each log entry is a line of the form:
 * error,output,setpoint
 */
class LoggingPIDOutput : public PIDOutput {
public:
    // No cleanup on any of these args is done during this classes' destructor
    LoggingPIDOutput(PIDOutput *out, UDPLogger *logger, MonoEncoder *encoder);
    virtual ~LoggingPIDOutput();

    // MUST call this BEFORE PIDController using this LoggingPIDOutput is Enabled
    void setController(PIDController *controller);
    virtual void PIDWrite(float output);

private:
    PIDController *controller;
    PIDOutput *out;
    UDPLogger *logger;
    MonoEncoder *encoder;
};


#endif
